package weltkante.scanner

import collection.mutable.{ArrayBuffer, Stack}
import weltkante.ScalaUtils._

//import util.control.Breaks._

import STokenFactories._

class Scanner(val mSource: Iterable[Char]) {
  val kStateMask = 0x07
  val kStateNone = 0x00
  val kStateGroup = 0x01
  val kStateIndex = 0x02
  val kStateContext = 0x03
  val kStateHint = 0x04
  val kStateString = 0x05

  val kAllowCloseHint = 0x10
  val kAllowUnsafeSymbol = 0x20

  val mInput = mSource.iterator
  var mBuffer = new StringBuilder
  var mOffset = 0
  var mTokens = new ArrayBuffer[SToken]
  var mStack = new Stack[Int]
  var mState = kAllowUnsafeSymbol;
  var mDidPushEnd = false

  def getToken(lookAhead: Int): SToken = {
    assert(lookAhead >= 0)
    ensureToken(lookAhead)
    mTokens(lookAhead)
  }

  def popToken: SToken = {
    ensureToken(0)
    val token = mTokens.head
    mTokens.remove(0)
    token
  }

  def popToken(count: Int) {
    assert(count >= 0)
    ensureToken(count - 1)
    mTokens.remove(0, count)
  }

  def ensureToken(lookAhead: Int) {
    assert(lookAhead >= 0)
    while (lookAhead >= mTokens.length)
      if (scanToken.isInstanceOf[TEnd] && mStack.length != 0)
        throw new ScannerException("Unmatched group.")
  }

  def popState(state: Int) {
    if ((mState & kStateMask) != state)
      throw new ScannerException("Mismatched group.")

    mState = mStack.pop()
  }

  def pushState(state: Int) {
    mStack.push(mState)
    mState = state
  }

  def scanToken: SToken = {
    assert(mOffset == 0)

    if (!hasInput) {
      if (mDidPushEnd)
        throw new ScannerEndOfInputException

      mDidPushEnd = true
      return pushToken[TEnd]
    }

    popChar match {
      case ' ' => pushToken[TSpace]
      case '\t' => pushToken[TTab]
      case '\n' => pushToken[TNewline]
      case '\r' => scanNewline2
      case '(' => scanOpenGroup
      case ')' => scanCloseGroup
      case '[' => scanOpenIndex
      case ']' => scanCloseIndex
      case '{' => scanOpenContext
      case '}' => scanCloseContext
      case '<' => scanOpenHintOrSymbol
      case '>' => scanCloseHintOrSymbol
      case '"' => scanStringHead
      case '#' => scanComment
      case '$' => scanBinary
      case '-' => scanSymbolOrCloseHint
      case '!' | '%' | '&' | '\'' | '*' | '+' | ',' | '.' | '/' | '|' | ';' |
           '=' | '?' | '@' | '\\' | '^' | '`' | '|' | '~' | ':' => pushToken[TSymbol]
      case ch if '0' <= ch && ch <= '9' => scanNumber
      case ch if isIdentHead(ch) => scanIdent
      case ch => throw new ScannerException("Unexpected input.")
    }
  }

  def scanNewline2 = {
    if (hasInput && peekChar == '\n')
      popChar

    pushToken[TNewline]
  }

  def isDigit(ch: Char): Boolean = '0' <= ch && ch <= '9'

  // TODO: I'm not yet sure whether '_' should be a symbol or identifier.

  def isIdentHead(ch: Char): Boolean = {
    ch == '_' ||
      'a' <= ch && ch <= 'z' ||
      'A' <= ch && ch <= 'Z'
  }

  def isIdentTail(ch: Char): Boolean = {
    ch == '_' ||
      'a' <= ch && ch <= 'z' ||
      'A' <= ch && ch <= 'Z' ||
      '0' <= ch && ch <= '9'
  }

  def scanOpenGroup: SToken = {
    pushState(kStateGroup | kAllowUnsafeSymbol)
    pushToken[TOpenGroup]
  }

  def scanCloseGroup: SToken = {
    popState(kStateGroup)
    pushToken[TCloseGroup]
  }

  def scanOpenIndex: SToken = {
    pushState(kStateIndex | kAllowUnsafeSymbol)
    return pushToken[TOpenIndex]
  }

  def scanCloseIndex: SToken = {
    popState(kStateIndex)
    return pushToken[TCloseIndex]
  }

  def scanOpenContext: SToken = {
    pushState(kStateContext | kAllowUnsafeSymbol)
    pushToken[TOpenContext]
  }

  def scanCloseContext: SToken = {
    popState(kStateContext)
    if ((mState & kStateMask) == kStateString)
      scanStringTail
    else
      pushToken[TCloseContext]
  }

  def scanOpenHintOrSymbol: SToken = {
    if ((mState & kAllowUnsafeSymbol) != 0) {
      val index = mTokens.length
      var count = 1
      pushToken[TSymbol]

      while (hasInput && peekChar == '<') {
        popChar
        pushToken[TSymbol]
        count += 1
      }

      forever {
        mStack.push(mState)
        mState &= ~kAllowUnsafeSymbol
        mState |= kAllowCloseHint

        var token = scanToken
        while (!(token.isInstanceOf[TClose] || token.isInstanceOf[TEnd]))
          token = scanToken

        if (token.isInstanceOf[TCloseHint]) {
          count -= 1

          val fixedToken = new TOpenHint
          fixedToken.fragment = mTokens(index + count).fragment
          mTokens(index + count) = fixedToken

          if (count == 0)
            return token
        } else {
          mState = mStack.pop()

          if ((mState & kAllowUnsafeSymbol) == 0)
            throw new ScannerException("Found more than one unsafe symbol in the same context.")

          mState &= ~kAllowUnsafeSymbol
          return token
        }
      }
    } else {
      pushState(kStateHint | kAllowCloseHint)
      return pushToken[TOpenHint]
    }
  }

  def scanCloseHintOrSymbol: SToken = {
    if ((mState & kAllowCloseHint) != 0) {
      mState = mStack.pop()
      return pushToken[TCloseHint]
    } else if ((mState & kAllowUnsafeSymbol) != 0) {
      mState &= ~kAllowUnsafeSymbol
      var token = pushToken[TSymbol]
      while (hasInput && peekChar == '>') {
        popChar
        token = pushToken[TSymbol]
      }
      return token
    } else {
      throw new ScannerException("Found more than one unsafe symbol in the same context.")
    }
  }

  def scanSymbolOrCloseHint: SToken = {
    // Workaround for the '->' operator.
    // Requires that '-' cannot appear as suffix operator in a hint-group.
    if (hasInput && peekChar == '>')
      popChar
    return pushToken[TSymbol]
  }

  def scanNumber: SToken = {
    var real = false

    while (hasInput && isDigit(peekChar)) popChar

    if (hasInput(1) && peekChar == '.' && isDigit(peekChar(1))) {
      real = true
      popChar
      popChar
      while (hasInput && isDigit(peekChar)) popChar
    }

    if (hasInput(1)) {
      var ch = peekChar(0)
      if (ch == 'e' || ch == 'E') {
        var signed = false
        ch = peekChar(1)
        if ((ch == '+' || ch == '-') && hasInput(2)) {
          signed = true
          ch = peekChar(2)
        }
        if (isDigit(ch)) {
          real = true
          popChar
          popChar
          if (signed) popChar
          while (hasInput && isDigit(peekChar)) popChar
        }
      }
    }

    return if (real) pushToken[TFloat] else pushToken[TInteger]
  }

  def scanIdent: SToken = {
    while (hasInput && isIdentTail(peekChar)) popChar
    return pushToken[TIdent]
  }

  def scanStringHead: SToken = {
    forever {
      popChar match {
        case '\n' =>
          throw new ScannerException("Unexpected newline in string.")
        case '"' =>
          return pushToken[TString]
        case '{' =>
          pushState(kStateString)
          pushState(kStateContext | kAllowUnsafeSymbol)
          return pushToken[TStringHead]
      }
    }
  }

  def scanStringTail: SToken = {
    forever {
      popChar match {
        case '\n' =>
          throw new ScannerException("Unexpected newline in string.")
        case '"' =>
          popState(kStateString)
          return pushToken[TStringTail]
        case '{' =>
          pushState(kStateContext | kAllowUnsafeSymbol)
          return pushToken[TStringPart]
      }
    }
  }

  def scanBinary: SToken = {
    if (popChar != '[')
      throw new ScannerException("Expected '['.")

    val binary = new ArrayBuffer[Byte]
    val number = new ArrayBuffer[Byte]
    var halfByte = -1

    forever {
      popChar match {
        case ']' =>
          if (halfByte >= 0)
            throw new ScannerException("Incomplete byte.")
          binary ++= number.reverseIterator
          val data = pushToken[TData]
          data.content = binary.toArray
          return data
        case ' ' | '\n' | '\t' =>
          if (halfByte >= 0)
            throw new ScannerException("Incomplete byte.")
          binary ++= number.reverseIterator
          number.clear()
        case ch if '0' <= ch && ch <= '9' =>
          if (halfByte < 0) {
            halfByte = ch - '0'
          } else {
            number += ((halfByte << 4) + (ch - '0')).toByte
            halfByte = -1
          }
        case ch if 'a' <= ch && ch <= 'f' =>
          if (halfByte < 0) {
            halfByte = ch - 'a'
          } else {
            number += ((halfByte << 4) + (ch - 'a')).toByte
            halfByte = -1
          }
        case ch if 'A' <= ch && ch <= 'F' =>
          if (halfByte < 0) {
            halfByte = ch - 'A'
          } else {
            number += ((halfByte << 4) + (ch - 'A')).toByte
            halfByte = -1
          }
        case _ =>
          throw new ScannerException("Unexpected input.")
      }
    }
  }

  def scanComment: SToken = {
    while (popChar != '\n') {}
    return pushToken[TComment]
  }

  def pushToken[TToken <: SToken](implicit factory: STokenFactory[TToken]): TToken = {
    val token = factory.create()
    token.fragment = mBuffer.substring(0, mOffset)
    mBuffer.delete(0, mOffset)
    mOffset = 0
    mTokens.append(token)
    return token
  }

  def hasInput: Boolean = hasInput(0)

  def hasInput(offset: Int): Boolean = {
    assert(offset >= 0)
    while (mBuffer.length - mOffset <= offset) {
      if (!mInput.hasNext)
        return false
      val next = mInput.next()
      if (next < 0)
        throw new Exception("huh?")
      mBuffer.append(next)
    }
    return true
  }

  def peekChar: Char = peekChar(0)

  def peekChar(offset: Int): Char = {
    assert(offset >= 0)
    if (!hasInput(offset))
      throw new ScannerEndOfInputException
    return mBuffer(mOffset + offset)
  }

  def popChar: Char = {
    if (mBuffer.length == mOffset) {
      if (!mInput.hasNext)
        throw new ScannerEndOfInputException

      val next = mInput.next()
      if (next < 0) throw new Exception("huh?")
      mBuffer.append(next)
      mOffset += 1
      return next
    } else {
      val next = mBuffer(mOffset)
      mOffset += 1
      return next
    }
  }
}
